/*
* PyUpnpBrowser
* Copyright (C) 2008 Instituto Nokia de Tecnologia
* Author: Renato Chencarek <renato.chencarek@openbossa.org>
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*
* Additional permission under GNU GPL version 3 section 7
*
* If you modify this Program, or any covered work, by linking or combining it
* with LibUPnPBrowser or Canola2 and its core components (or a modified
* version of any of those), containing parts covered by the terms of Instituto
* Nokia de Tecnologia End User Software Agreement, the licensors of this
* Program grant you additional permission to convey the resulting work.
*/

#include <Python.h>
#include "structmember.h"

#include <libupnpbrowser/upnp-browser.h>
#include <cybergarage/upnp/ccontrolpoint.h>

#include "upnp-browser-module.h"
#include "upnp-device-module.h"
#include "upnp-item-module.h"

static PyThreadState *myThreadState = NULL;

static void
Browser_dealloc(Browser* self)
{
    if (self->browser_priv)
    {
        upnp_browser_stop(self->browser_priv);
        upnp_browser_destroy(self->browser_priv);
    }
    self->ob_type->tp_free((PyObject*)self);
}

static PyObject *
Browser_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
    Browser *self;

    self = (Browser*)type->tp_alloc(type, 0);
    return (PyObject *)self;
}

static int
Browser_init(Browser *self, PyObject *args, PyObject *kwds)
{
    self->browser_priv = upnp_browser_new();
    if (self->browser_priv)
        return 0;
    else
        return -1;
}

static PyObject *
Browser_start(Browser *self)
{
    upnp_browser_start(self->browser_priv);
    Py_RETURN_NONE;
}

static PyObject *
Browser_stop(Browser *self)
{
    upnp_browser_stop(self->browser_priv);
    Py_RETURN_NONE;
}

static PyObject *
_get_device(upnp_browser_t *browser_priv, const char* udn)
{

    Device *device = (Device*)PyObject_CallObject((PyObject *)&DeviceType, NULL);

    gchar *dev_name, **sort_array, **sort_itr;
    GError *error = NULL;

    dev_name = upnp_browser_get_device_name(browser_priv, udn);
    sort_array = upnp_browser_get_sort_caps(browser_priv, udn, &error);

    if( error )
    {
        fprintf(stderr, "\nERROR: PyUPnPBrowser: _get_device: \n");
        fprintf(stderr, "upnp_browser_get_sort_caps: %s\n", error->message);
        g_error_free (error);
    }


    Py_DECREF(device->name);
    device->name = Py_BuildValue("s", dev_name);

    Py_DECREF(device->udn);
    device->udn = Py_BuildValue("s", udn);

    g_free (dev_name);

    if (sort_array)
    {
        Py_DECREF(device->sort_caps);
        device->sort_caps = PyList_New(0);

        if (!error)
            for (sort_itr = sort_array; *sort_itr != NULL; sort_itr++)
            {
                PyObject *str = Py_BuildValue("s", *sort_itr);
                PyList_Append(device->sort_caps, str);
                Py_DECREF(str);
            }
        else
            g_error_free (error);

        g_strfreev (sort_array);
    }

    return (PyObject*)device;
}

static PyObject *
Browser_get_device(Browser *self, PyObject *args, PyObject *keywds)
{
    Device *device = NULL;
    static char *kwlist[] = {"udn", NULL};

    char* udn;

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "s", kwlist, &udn))
        return NULL;

    device =  (Device*)_get_device(self->browser_priv, udn);

    Py_DECREF(device->browser_priv);
    Py_INCREF(self);
    device->browser_priv = (PyObject*)self;

    return (PyObject*)device;
}

static PyObject *
Browser_list_devices(Browser *self)
{
    GSList *devices, *itr;
    PyObject* list = PyList_New(0);

    devices = upnp_browser_get_devices(self->browser_priv);
    for (itr = devices; itr != NULL; itr = itr->next) {
        char *udn = itr->data;

        Device *device = (Device*) _get_device(self->browser_priv, udn);

        Py_DECREF(device->browser_priv);
        Py_INCREF(self);
        device->browser_priv = (PyObject*)self;

        PyList_Append(list, (PyObject*)device);
        Py_DECREF(device);

        g_free (udn);
    }

    g_slist_free (devices);

    return list;
}

static const gchar *
_status_to_str (CgUpnpDeviceStatus status)
{
    static const gchar *texts[] = {
        [UPNP_BROWSER_DEVICE_STATUS_ADDED] = "ADDED",
        [UPNP_BROWSER_DEVICE_STATUS_REMOVED] = "REMOVED",
        [UPNP_BROWSER_DEVICE_STATUS_UPDATED] = "UPDATED",
        [UPNP_BROWSER_DEVICE_STATUS_INVALID] = "INVALID",
    };
    if (status == UPNP_BROWSER_DEVICE_STATUS_ADDED ||
        status == UPNP_BROWSER_DEVICE_STATUS_REMOVED ||
        status == UPNP_BROWSER_DEVICE_STATUS_UPDATED ||
        status == UPNP_BROWSER_DEVICE_STATUS_INVALID)
        return texts[status];
    else
        return NULL;
}

static void
_device_listener (upnp_browser_t *browser,
                  const char *udn,
                  upnp_browser_device_status_t status,
                  gpointer user_data)
{
    Browser* self = (Browser*) user_data;
    Device* device = (Device*)_get_device(browser, udn);

    PyObject *arglist;
    PyObject *result = NULL;

    Py_INCREF(device);
    if (PyCallable_Check(self->listener))
    {
        PyEval_AcquireLock();
        PyThreadState_Swap(myThreadState);

        arglist = Py_BuildValue("(OsO)", device,
                                _status_to_str(status),
                                self->user_data);
        Py_INCREF(arglist);
        result = PyEval_CallObject(self->listener, arglist);
        Py_DECREF(arglist);

        PyThreadState_Swap(NULL);
        PyEval_ReleaseLock();
    }
    Py_DECREF(device);

    if (result != NULL)
    {
        Py_DECREF(result);
    }
}

static PyObject *
Browser_get_notification_fd(Browser *self)
{
    int fd = upnp_browser_get_device_notification_fd (self->browser_priv);
    if ( fd < 0 )
        Py_RETURN_NONE;

    return Py_BuildValue("i", fd);
}

static PyObject *
Browser_close_notification_fd(Browser *self, PyObject *args, PyObject *keywds)
{
    static char *kwlist[] = {"fd", NULL};

    int fd;

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "i", kwlist, &fd))
        return NULL;

    upnp_browser_remove_device_notification_fd (self->browser_priv, fd);

    Py_RETURN_NONE;
}


static PyObject *
Browser_set_listener(Browser *self, PyObject *args, PyObject *keywds)
{
    static char *kwlist[] = {"listener", "user_data", NULL};

    PyObject *listener  = NULL;
    PyObject *user_data = NULL;
    PyObject *tmp;

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "|OO", kwlist,
                                     &listener, &user_data))
        return NULL;

    if (listener)
    {
        if (!PyCallable_Check(listener))
        {
            PyErr_SetString(PyExc_TypeError, "parameter must be callable");
            return NULL;
        }

        tmp = self->listener;
        Py_INCREF(listener);
        Py_DECREF(tmp);
        self->listener = listener;
    }

    if (user_data)
    {
        tmp = self->user_data;
        Py_INCREF(user_data);
        Py_DECREF(tmp);
        self->user_data = user_data;
    }


    if (self->listener)
    {
        if (myThreadState == NULL)
        {
            PyThreadState *mainThreadState;
            PyInterpreterState *mainInterpreterState;

            mainThreadState = PyThreadState_Get();
            mainInterpreterState = mainThreadState->interp;
            myThreadState = PyThreadState_New(mainInterpreterState);
        }

        upnp_browser_set_device_listener(self->browser_priv,
                                         _device_listener, self, NULL);
    }
    else
        upnp_browser_set_device_listener(self->browser_priv,
                                         NULL, NULL, NULL);

    Py_RETURN_NONE;
}

static PyMemberDef Browser_members[] = {
    {NULL}  /* Sentinel */
};

static PyMethodDef Browser_methods[] = {
    {"start_search", (PyCFunction)Browser_start, METH_NOARGS,
     "Start the search for UPnP Servers."},

    {"stop_search", (PyCFunction)Browser_stop, METH_NOARGS,
     "Stop the search for UPnP Servers."},

    {"list_devices", (PyCFunction)Browser_list_devices, METH_NOARGS,
     "List all UPnP servers found."},

    {"get_device", (PyCFunction)Browser_get_device, METH_VARARGS|METH_KEYWORDS,
     "Return the UPnP Device with udn."},

    {"set_listener", (PyCFunction)Browser_set_listener, METH_VARARGS|METH_KEYWORDS,
     "Set Device listener."},

    {"notification_fd", (PyCFunction)Browser_get_notification_fd, METH_NOARGS,
     "Get file descriptor to Devices notification."},

    {"close_fd", (PyCFunction)Browser_close_notification_fd, METH_VARARGS|METH_KEYWORDS,
     "Close file descriptor to Devices notification."},

    { NULL, NULL, 0, NULL }
};

PyTypeObject BrowserType = {
    PyObject_HEAD_INIT(NULL)
    0,                               /*ob_size*/
    "Browser",                   /*tp_name*/
    sizeof(Browser),             /*tp_basicsize*/
    0,                               /*tp_itemsize*/
    (destructor)Browser_dealloc, /*tp_dealloc*/
    0,                               /*tp_print*/
    0,                               /*tp_getattr*/
    0,                               /*tp_setattr*/
    0,                               /*tp_compare*/
    0,                               /*tp_repr*/
    0,                               /*tp_as_number*/
    0,                               /*tp_as_sequence*/
    0,                               /*tp_as_mapping*/
    0,                               /*tp_hash */
    0,                               /*tp_call*/
    0,                               /*tp_str*/
    0,                               /*tp_getattro*/
    0,                               /*tp_setattro*/
    0,                               /*tp_as_buffer*/
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
    "Browser objects",      /* tp_doc */
    0,		                    /* tp_traverse */
    0,		                    /* tp_clear */
    0,		                    /* tp_richcompare */
    0,		                    /* tp_weaklistoffset */
    0,		                    /* tp_iter */
    0,		                    /* tp_iternext */
    Browser_methods,        /* tp_methods */
    Browser_members,        /* tp_members */
    0,                          /* tp_getset */
    0,                          /* tp_base */
    0,                          /* tp_dict */
    0,                          /* tp_descr_get */
    0,                          /* tp_descr_set */
    0,                          /* tp_dictoffset */
    (initproc)Browser_init, /* tp_init */
    0,                          /* tp_alloc */
    Browser_new,            /* tp_new */
};

static PyObject *
upnpbrowser_init(PyObject *self, PyObject *args)
{
    upnp_browser_init();
    PyEval_InitThreads();

    Py_RETURN_NONE;
}

static PyObject *
upnpbrowser_quit(PyObject *self, PyObject *args)
{
    upnp_browser_quit();

    Py_RETURN_NONE;
}

static PyObject *
upnpbrowser_read_data(Browser *self, PyObject *args, PyObject *keywds)
{
    int fd, size;
    char udn[256];
    char status[64];

    static char *kwlist[] = {"fd", NULL};

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "i", kwlist, &fd))
        return NULL;

    read(fd, &size, sizeof(int));
    read(fd, &udn[0], size);
    udn[size] = 0;

    read(fd, &size, sizeof(int));
    read(fd, &status[0], size);
    status[size] = 0;

    return Py_BuildValue("(ss)", udn, status);
}

static PyMethodDef module_methods[] = {
    {"init", (PyCFunction)upnpbrowser_init, METH_NOARGS,
     "Init UPnP Engine."},
    {"quit", (PyCFunction)upnpbrowser_quit, METH_NOARGS,
     "Quit UPnP Engine."},
    {"read_data", (PyCFunction)upnpbrowser_read_data, METH_VARARGS|METH_KEYWORDS,
     "Read data (udn, status) from FD."},
  {NULL}
};

void
initupnpbrowser(void)
{
    PyObject* m;

    if (PyType_Ready(&BrowserType) < 0)
        return;

    if (PyType_Ready(&DeviceType) < 0)
        return;

    if (PyType_Ready(&ItemType) < 0)
        return;


    m = Py_InitModule3("upnpbrowser", module_methods,
                       "LibUPnPBrowser python module.");

    if (m == NULL)
      return;

    Py_INCREF(&BrowserType);
    PyModule_AddObject(m, "Browser", (PyObject *)&BrowserType);

    Py_INCREF(&DeviceType);
    PyModule_AddObject(m, "Device", (PyObject *)&DeviceType);

    Py_INCREF(&ItemType);
    PyModule_AddObject(m, "Item", (PyObject *)&ItemType);

    upnpbrowser_init(NULL, NULL);
}
