/*
* PyUpnpBrowser
* Copyright (C) 2008 Instituto Nokia de Tecnologia
* Author: Renato Chencarek <renato.chencarek@openbossa.org>
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*
* Additional permission under GNU GPL version 3 section 7
*
* If you modify this Program, or any covered work, by linking or combining it
* with LibUPnPBrowser or Canola2 and its core components (or a modified
* version of any of those), containing parts covered by the terms of Instituto
* Nokia de Tecnologia End User Software Agreement, the licensors of this
* Program grant you additional permission to convey the resulting work.
*/

#include <Python.h>
#include "structmember.h"

#include <libupnpbrowser/upnp-browser.h>
#include <cybergarage/upnp/ccontrolpoint.h>

#include "upnp-device-module.h"
#include "upnp-browser-module.h"
#include "upnp-item-module.h"

static void
Device_dealloc(Device* self)
{
    Py_XDECREF(self->name);
    Py_XDECREF(self->udn);
    Py_XDECREF(self->sort_caps);
    Py_XDECREF(self->browser_priv);

    self->ob_type->tp_free((PyObject*)self);
}

static PyObject *
Device_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
    Device *self;

    self = (Device *)type->tp_alloc(type, 0);
    if (self != NULL)
    {
        self->name = PyString_FromString("");
        if (!self->name)
        {
            Py_DECREF(self);
            return NULL;
        }

        self->udn = PyString_FromString("");
        if (!self->udn)
        {
            Py_DECREF(self);
            return NULL;
        }

        self->sort_caps = PyString_FromString("");
        if (!self->sort_caps)
        {
            Py_DECREF(self);
            return NULL;
        }

        Py_INCREF(Py_None);
        self->browser_priv = Py_None;
    }

    return (PyObject *)self;
}

static int
Device_init(Device *self, PyObject *args, PyObject *kwds)
{
    PyObject *name=NULL, *udn=NULL, *sort_caps=NULL, *browser_priv=NULL, *tmp;

    static char *kwlist[] = {"name", "udn", "sort_caps", "browser_priv", NULL};

    if (! PyArg_ParseTupleAndKeywords(args, kwds, "|OOOO", kwlist,
                                      &name, &udn, &sort_caps, &browser_priv))
        return -1;

    if (name)
    {
        tmp = self->name;
        Py_INCREF(name);
        self->name = name;
        Py_DECREF(tmp);
    }

    if (udn)
    {
        tmp = self->udn;
        Py_INCREF(udn);
        self->udn = udn;
        Py_DECREF(tmp);
    }

    if (browser_priv)
    {
        tmp = self->browser_priv;
        Py_INCREF(browser_priv);
        self->browser_priv = browser_priv;
        Py_DECREF(tmp);
    }

    if (sort_caps)
    {
        tmp = self->sort_caps;
        Py_INCREF(sort_caps);
        self->sort_caps = sort_caps;
        Py_DECREF(tmp);
    }

    return 0;
}

static void
_process_data_foreach (const gchar *key,
                       const gchar *value,
                       PyObject *attributes)
{
    PyObject *tmp_key  = Py_BuildValue("s", key);
    PyObject *tmp_value = Py_BuildValue("s", value);

    PyDict_SetItem(attributes, tmp_key, tmp_value);

    Py_DECREF(tmp_key);
    Py_DECREF(tmp_value);
}

static void
_process_info_foreach (const char *key,
                       const GSList *value,
                       Item *item)
{
    const GSList *itr;
    PyObject *tmp;
    PyObject* contents = PyList_New(0);

    for (itr = value; itr != NULL; itr = itr->next)
    {
        PyObject* data;
        PyObject* attributes = PyDict_New();
        upnp_cds_info_t *cds = (upnp_cds_info_t*)itr->data;

        g_hash_table_foreach (cds->attributes,
                              (GHFunc)_process_data_foreach, attributes);

        tmp = Py_BuildValue("s", cds->contents);

        data = PyTuple_Pack(2, tmp, attributes);

        Py_DECREF(attributes);
        Py_DECREF(tmp);

        PyList_Append(contents, data);
        Py_DECREF(data);
    }

    tmp = Py_BuildValue("s", key);
    PyDict_SetItem(item->info, tmp, contents);
    Py_DECREF(tmp);
    Py_DECREF(contents);
}

static void
_process_info (Item *item, GHashTable *info)
{
    item->info = PyDict_New();
    g_hash_table_foreach (info, (GHFunc)_process_info_foreach, item);
}

static PyObject *
Device_list_folder(Device *self, PyObject *args, PyObject *keywds)
{
    Browser *browser = (Browser*)self->browser_priv;
    PyObject* list = PyList_New(0);

    GSList *items, *itr;
    GError *error = NULL;
    char *parent_id, *udn, *property_filter = NULL, *sort_filter = NULL;
    int start = 0, max_count = 0;

    static char *kwlist[] = {"id", "start", "max_count",
                             "property_filter", "sort_filter", NULL};


    if (!PyArg_ParseTupleAndKeywords(args, keywds, "s|iiss", kwlist,
                                     &parent_id, &start, &max_count,
                                     &property_filter, &sort_filter))
        return NULL;

    if (PyString_Check(self->udn))
        udn = PyString_AsString(self->udn);
    else
        Py_RETURN_NONE;

    error = NULL;
    items = upnp_browser_list_folder_complete (browser->browser_priv, udn,
                                               parent_id, start, max_count,
                                               property_filter, sort_filter,
                                               &error);
    if (error)
    {
        fprintf(stderr, "\nERROR: PyUPnPBrowser: Device_list_folder: \n");
        fprintf(stderr, "upnp_browser_list_folder_complete: %s\n", error->message);
        g_error_free (error);

        Py_RETURN_NONE;
    }

    for (itr = items; itr != NULL; itr = itr->next)
    {
        Item *item = (Item*)PyObject_CallObject((PyObject *)&ItemType, NULL);
        upnp_cds_object_t *object = itr->data;


        error = NULL;
        if(error)
        {
            fprintf(stderr, "\nERROR: PyUPnPBrowser: Device_list_folder: \n");
            fprintf(stderr, "upnp_browser_get_object: %s\n", error->message);

            g_error_free (error);
            error = NULL;
            continue;
        }

        if(!object)
        {
            fprintf(stderr, "\nERROR: PyUPnPBrowser: Device_list_folder: \n");
            fprintf(stderr, "upnp_browser_get_object: object==NULL: %s\n",
                    object->id);
            continue;
        }

        Py_DECREF(item->type);
        item->type = Py_BuildValue("i", object->type);

        Py_DECREF(item->klass);
        item->klass = Py_BuildValue("s", object->klass);

        Py_DECREF(item->parent_id);
        item->parent_id = Py_BuildValue("s", object->parent_id);

        Py_DECREF(item->id);
        item->id = Py_BuildValue("s", object->id);

        Py_DECREF(item->title);
        item->title = Py_BuildValue("s", object->title);

        Py_DECREF(item->creator);
        item->creator = Py_BuildValue("s", object->creator);


        Py_DECREF(item->restricted);
        if (object->restricted)
            item->restricted = Py_BuildValue("i", 1);
        else
            item->restricted = Py_BuildValue("i", 0);


        Py_DECREF(item->info);
        _process_info (item, object->info);

        upnp_cds_object_destroy(object);

        PyList_Append(list, (PyObject*)item);
        Py_DECREF(item);
    }

    g_slist_free (items);
    return list;
}

static PyObject*
Device_list_services(Device *self)
{
    Browser *browser = (Browser*)self->browser_priv;
    GSList *services, *itr;
    PyObject* list = PyList_New(0);

    char* udn = PyString_AsString(self->udn);

    services = upnp_browser_get_device_services(browser->browser_priv, udn);
    for (itr = services; itr != NULL; itr = itr->next)
    {
        char *service_name = itr->data;

        PyObject *str = Py_BuildValue("s", service_name);
        PyList_Append(list, str);
        Py_DECREF(str);

        g_free (service_name);
    }

    g_slist_free (services);

    return list;
}


static PyObject*
Device_Repr (PyObject *obj)
{
    PyObject *ret, *format, *template;;
    Device *self = (Device*) obj;


    format = Py_BuildValue("s", "< Device: %s\n  udn: %s\n  sort_caps: %s >");
    template = Py_BuildValue("(OOO)", self->name, self->udn, self->sort_caps);
    ret = PyString_Format(format, template);

    Py_DECREF(format);
    Py_DECREF(template);

    return ret;
}

static int
Device_Compare (PyObject *o1, PyObject *o2)
{
    Device *d1 = (Device*)o1;
    Device *d2 = (Device*)o2;

    return strcmp(PyString_AsString(d1->udn),
                  PyString_AsString(d2->udn));
}

static PyMemberDef Device_members[] = {
    {"name", T_OBJECT_EX, offsetof(Device, name), 0,
     "device name"},
    {"udn", T_OBJECT_EX, offsetof(Device, udn), 0,
     "device udn"},
    {"sort_caps", T_OBJECT_EX, offsetof(Device, sort_caps), 0,
     "sort capabilities description"},

    {NULL}
};

static PyMethodDef Device_methods[] = {
    {"list_folder", (PyCFunction)Device_list_folder, METH_VARARGS|METH_KEYWORDS,
     "List all items inside the folder with id (id == 0 for Root)."},

    {"list_services", (PyCFunction)Device_list_services, METH_NOARGS,
     "Return the UPnP Services of the Device."},

    { NULL, NULL, 0, NULL }
};

PyTypeObject DeviceType = {
    PyObject_HEAD_INIT(NULL)
    0,                               /*ob_size*/
    "Device",                   /*tp_name*/
    sizeof(Device),             /*tp_basicsize*/
    0,                               /*tp_itemsize*/
    (destructor)Device_dealloc, /*tp_dealloc*/
    0,                               /*tp_print*/
    0,                               /*tp_getattr*/
    0,                               /*tp_setattr*/
    (cmpfunc)Device_Compare,         /*tp_compare*/
    (reprfunc)Device_Repr,           /*tp_repr*/
    0,                               /*tp_as_number*/
    0,                               /*tp_as_sequence*/
    0,                               /*tp_as_mapping*/
    0,                               /*tp_hash */
    0,                               /*tp_call*/
    0,                               /*tp_str*/
    0,                               /*tp_getattro*/
    0,                               /*tp_setattro*/
    0,                               /*tp_as_buffer*/
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
    "Device objects",      /* tp_doc */
    0,		                    /* tp_traverse */
    0,		                    /* tp_clear */
    0,		                    /* tp_richcompare */
    0,		                    /* tp_weaklistoffset */
    0,		                    /* tp_iter */
    0,		                    /* tp_iternext */
    Device_methods,        /* tp_methods */
    Device_members,        /* tp_members */
    0,                          /* tp_getset */
    0,                          /* tp_base */
    0,                          /* tp_dict */
    0,                          /* tp_descr_get */
    0,                          /* tp_descr_set */
    0,                          /* tp_dictoffset */
    (initproc)Device_init, /* tp_init */
    0,                          /* tp_alloc */
    Device_new,            /* tp_new */
};
