/*
* PyUpnpBrowser
* Copyright (C) 2008 Instituto Nokia de Tecnologia
* Author: Renato Chencarek <renato.chencarek@openbossa.org>
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*
* Additional permission under GNU GPL version 3 section 7
*
* If you modify this Program, or any covered work, by linking or combining it
* with LibUPnPBrowser or Canola2 and its core components (or a modified
* version of any of those), containing parts covered by the terms of Instituto
* Nokia de Tecnologia End User Software Agreement, the licensors of this
* Program grant you additional permission to convey the resulting work.
*/

#include <Python.h>
#include "structmember.h"

#include <libupnpbrowser/upnp-browser.h>
#include <cybergarage/upnp/ccontrolpoint.h>

#include "upnp-item-module.h"

static void
Item_dealloc(Item* self)
{
    Py_DECREF(self->type);
    Py_DECREF(self->klass);
    Py_DECREF(self->parent_id);

    Py_DECREF(self->id);
    Py_DECREF(self->title);
    Py_DECREF(self->creator);
    Py_DECREF(self->restricted);

    Py_DECREF(self->info);

    self->ob_type->tp_free((PyObject*)self);
}

static PyObject *
Item_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
    Item *self;

    self = (Item *)type->tp_alloc(type, 0);
    if (self != NULL)
    {
        self->type = Py_BuildValue("i", 0);

        Py_INCREF(Py_None);
        self->klass = Py_None;

        Py_INCREF(Py_None);
        self->parent_id = Py_None;

        Py_INCREF(Py_None);
        self->id = Py_None;

        Py_INCREF(Py_None);
        self->title = Py_None;

        Py_INCREF(Py_None);
        self->creator = Py_None;

        self->restricted = Py_BuildValue("i", 0);

        Py_INCREF(Py_None);
        self->info = Py_None;
    }

    return (PyObject *)self;
}

static int
Item_init(Item *self, PyObject *args, PyObject *kwds)
{
    PyObject* tmp;

    PyObject* type = NULL;
    PyObject* klass = NULL;
    PyObject* parent_id = NULL;

    PyObject* id = NULL;
    PyObject* title = NULL;
    PyObject* creator = NULL;
    PyObject* restricted = NULL;

    PyObject* info = NULL;

    static char *kwlist[] = {"type", "klass", "parent_id", "id", "title", "creator", "restricted", "info", NULL};

    if ( !PyArg_ParseTupleAndKeywords(args, kwds, "|OOOOOOOO", kwlist,
                                      &type, &klass, &parent_id,
                                      &id, &title, &creator,
                                      &restricted, &info) )
        return -1;

    if (type)
    {
        tmp = self->type;
        Py_INCREF(type);
        self->type = type;
        Py_DECREF(tmp);
    }

    if (klass)
    {
        tmp = self->klass;
        Py_INCREF(klass);
        self->klass = klass;
        Py_DECREF(tmp);
    }

    if (parent_id)
    {
        tmp = self->parent_id;
        Py_INCREF(parent_id);
        self->parent_id = parent_id;
        Py_DECREF(tmp);
    }

    if (id)
    {
        tmp = self->id;
        Py_INCREF(id);
        self->id = id;
        Py_DECREF(tmp);
    }

    if (title)
    {
        tmp = self->title;
        Py_INCREF(title);
        self->title = title;
        Py_DECREF(tmp);
    }

    if (creator)
    {
        tmp = self->creator;
        Py_INCREF(creator);
        self->creator = creator;
        Py_DECREF(tmp);
    }

    if (restricted)
    {
        tmp = self->restricted;
        Py_INCREF(restricted);
        self->restricted = restricted;
        Py_DECREF(tmp);
    }

    if (info)
    {
        tmp = self->info;
        Py_INCREF(info);
        self->info = info;
        Py_DECREF(tmp);
    }

    return 0;
}

static PyObject *
Item_get_info(Item *self, PyObject *args, PyObject *kwds)
{
    PyObject* key = NULL;
    PyObject* res_list = NULL;
    PyObject* res_tuple = NULL;
    PyObject* ret;

    static char *kwlist[] = {"key", NULL};

    if ( !PyArg_ParseTupleAndKeywords(args, kwds, "O", kwlist,
                                      &key) )
        Py_RETURN_NONE;

    if (!key)
        Py_RETURN_NONE;

    if (!PyDict_Contains(self->info, key))
        Py_RETURN_NONE;

    res_list = PyDict_GetItem(self->info, key);

    res_tuple = PyList_GetItem(res_list, 0);

    ret = PyTuple_GetItem(res_tuple, 0);

    Py_INCREF(ret);

    return ret;
}

static PyObject *
Item_get_attribute(Item *self, PyObject *args, PyObject *kwds)
{
    PyObject* key = NULL;

    PyObject* res_key   = NULL;
    PyObject* res_tuple = NULL;
    PyObject* res_dict  = NULL;
    PyObject* res_list  = NULL;

    PyObject* ret;

    static char *kwlist[] = {"key", NULL};

    if ( !PyArg_ParseTupleAndKeywords(args, kwds, "O", kwlist,
                                      &key) )
        Py_RETURN_NONE;

    if (!key)
        Py_RETURN_NONE;

    res_key = Py_BuildValue("s", "res");
    if (!PyDict_Contains(self->info, res_key))
    {
        Py_DECREF(res_key);
        Py_RETURN_NONE;
    }

    res_list = PyDict_GetItem(self->info, res_key);
    Py_DECREF(res_key);

    res_tuple = PyList_GetItem(res_list, 0);

    res_dict = PyTuple_GetItem(res_tuple, 1);

    if (!PyDict_Contains(res_dict, key))
        Py_RETURN_NONE;

    ret = PyDict_GetItem(res_dict, key);

    Py_INCREF(ret);
    return ret;
}

static PyObject *
Item_urls(Item *self, PyObject *args, PyObject *kwds)
{
    PyObject* res_key = NULL;
    PyObject* res_list = NULL;

    PyObject* mimetypes = NULL;
    PyObject* ret = PyList_New(0);

    int i, j;
    char* hugestr;

    static char *kwlist[] = {"mimetypes", NULL};

    if ( (args || kwds) && !PyArg_ParseTupleAndKeywords(args, kwds, "O",
                                                        kwlist, &mimetypes) )
        Py_RETURN_NONE;

    res_key = Py_BuildValue("s", "res");
    if (!PyDict_Contains(self->info, res_key))
    {
        Py_DECREF(res_key);
        Py_RETURN_NONE;
    }

    res_list = PyDict_GetItem(self->info, res_key);
    Py_DECREF(res_key);

    PyObject *key = Py_BuildValue("s", "protocolInfo");
    for(i = 0; i < PyList_Size(res_list); i++)
    {
        PyObject *types = NULL;
        PyObject *tuple = PyList_GetItem(res_list, i);
        PyObject *info = PyTuple_GetItem(tuple, 1);

        if(!mimetypes)
        {
            PyList_Append(ret, PyTuple_GetItem(tuple, 0));
            continue;
        }

        if (!PyDict_Contains(info, key))
            continue;

        types = PyDict_GetItem(info, key);
        hugestr = PyString_AsString(types);

        for(j = 0; j < PyList_Size(mimetypes); j++)
        {
            char* mime = PyString_AsString(PyList_GetItem(mimetypes, j));
            if(strcasestr(hugestr, mime))
            {
                PyList_Append(ret, PyTuple_GetItem(tuple, 0));
                break;
            }
        }
    }
    Py_DECREF(key);


    Py_INCREF(ret);
    return ret;
}

static PyObject*
Item_Repr (PyObject *obj)
{
    PyObject *ret, *tmp, *template;
    PyObject* urls;
    Item *self = (Item*) obj;

    const char* format = "< Item: %s\n  type: %d\n\
  url: %s\n klass: %s\n  parent_id: %s\n  id: %s\n  creator: %s\n  restricted: %d\n\
  info: %s >";

    urls = Item_urls(self, NULL, NULL);
    tmp = Py_BuildValue("s", format);

    template = Py_BuildValue("(OOOOOOOOO)", self->title, self->type, urls,
                             self->klass, self->parent_id, self->id,
                             self->creator, self->restricted, self->info);

    ret = PyString_Format(tmp, template);

    Py_DECREF(urls);
    Py_DECREF(tmp);
    Py_DECREF(template);

    return ret;
}

static PyMemberDef Item_members[] = {
    {"type", T_OBJECT_EX, offsetof(Item, type), 0, "item type"},
    {"klass", T_OBJECT_EX, offsetof(Item, klass), 0, "item Class"},
    {"parent_id", T_OBJECT_EX, offsetof(Item, parent_id), 0, "parent id"},
    {"id", T_OBJECT_EX, offsetof(Item, id), 0, "id"},
    {"title", T_OBJECT_EX, offsetof(Item, title), 0, "title"},
    {"creator", T_OBJECT_EX, offsetof(Item, creator), 0, "creator"},
    {"restricted", T_OBJECT_EX, offsetof(Item, restricted), 0, "is restricted"},
    {"info", T_OBJECT_EX, offsetof(Item, info), 0, "information"},

    {NULL}  /* Sentinel */
};

static PyMethodDef Item_methods[] = {
    {"get_attribute", (PyCFunction)Item_get_attribute, METH_VARARGS|METH_KEYWORDS,
     "Get the attribute."},

    {"get_info",      (PyCFunction)Item_get_info, METH_VARARGS|METH_KEYWORDS,
     "Get the info list."},

    {"urls",      (PyCFunction)Item_urls, METH_VARARGS|METH_KEYWORDS,
     "URLS."},

    { NULL, NULL, 0, NULL }
};

PyTypeObject ItemType = {
    PyObject_HEAD_INIT(NULL)
    0,                               /*ob_size*/
    "Item",                   /*tp_name*/
    sizeof(Item),             /*tp_basicsize*/
    0,                               /*tp_itemsize*/
    (destructor)Item_dealloc, /*tp_dealloc*/
    0,                               /*tp_print*/
    0,                               /*tp_getattr*/
    0,                               /*tp_setattr*/
    0,                               /*tp_compare*/
    (reprfunc)Item_Repr,             /*tp_repr*/
    0,                               /*tp_as_number*/
    0,                               /*tp_as_sequence*/
    0,                               /*tp_as_mapping*/
    0,                               /*tp_hash */
    0,                               /*tp_call*/
    0,                               /*tp_str*/
    0,                               /*tp_getattro*/
    0,                               /*tp_setattro*/
    0,                               /*tp_as_buffer*/
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
    "Item objects",      /* tp_doc */
    0,		                    /* tp_traverse */
    0,		                    /* tp_clear */
    0,		                    /* tp_richcompare */
    0,		                    /* tp_weaklistoffset */
    0,		                    /* tp_iter */
    0,		                    /* tp_iternext */
    Item_methods,        /* tp_methods */
    Item_members,        /* tp_members */
    0,                          /* tp_getset */
    0,                          /* tp_base */
    0,                          /* tp_dict */
    0,                          /* tp_descr_get */
    0,                          /* tp_descr_set */
    0,                          /* tp_dictoffset */
    (initproc)Item_init, /* tp_init */
    0,                          /* tp_alloc */
    Item_new,            /* tp_new */
};
